# Generated by Django 3.2.9 on 2023-10-20 10:41

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('simo_sonos', '0001_initial'),
    ]

    operations = [
        migrations.RenameField(
            model_name='sonosplayer',
            old_name='master',
            new_name='slave_of',
        ),
    ]



# Generated by Django 3.2.9 on 2023-10-20 10:25

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='SonosPlayer',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('uid', models.CharField(db_index=True, max_length=200, unique=True)),
                ('name', models.CharField(db_index=True, max_length=200)),
                ('ip', models.GenericIPAddressField()),
                ('last_seen', models.DateTimeField(auto_now_add=True)),
                ('is_alive', models.BooleanField(default=True)),
                ('master', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='slaves', to='simo_sonos.sonosplayer')),
            ],
        ),
        migrations.CreateModel(
            name='SonosPlaylist',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('title', models.CharField(db_index=True, max_length=200)),
                ('item_id', models.CharField(max_length=200, unique=True)),
                ('player', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='simo_sonos.sonosplayer')),
            ],
            options={
                'unique_together': {('player', 'item_id')},
            },
        ),
    ]



import traceback
import sys
from simo.multimedia.controllers import BaseAudioPlayer
from .gateways import SONOSGatewayHandler
from .forms import SONOSPlayerConfigForm


class SONOSPlayer(BaseAudioPlayer):
    gateway_class = SONOSGatewayHandler
    config_form = SONOSPlayerConfigForm

    def unjoin(self):
        from simo_sonos.models import SonosPlayer
        sonos_player = SonosPlayer.objects.filter(
            id=self.component.config['sonos_device']
        ).first()
        if not sonos_player:
            return
        try:
            sonos_player.soco.unjoin()
        except:
            print(traceback.format_exc(), file=sys.stderr)

    def play_uri(self, uri, volume=None):
        if volume:
            assert 0 <= volume <= 100
        self.send({"play_uri": uri, 'volume': volume})

    def play_alert(self, val, volume=None):
        '''Val can be sound id or uri'''
        assert type(val) in (int, str)
        if volume:
            assert 0 <= volume <= 100
        self.send({"alert": val, 'volume': volume})

    def play_playlist(self, item_id, shuffle=True):
        from simo_sonos.models import SonosPlayer
        sonos_player = SonosPlayer.objects.filter(
            id=self.component.config['sonos_device']
        ).first()
        if not sonos_player:
            return
        for plst in sonos_player.soco.get_sonos_playlists():
            if plst.item_id == item_id:
                try:
                    sonos_player.soco.clear_queue()
                    sonos_player.soco.shuffle = shuffle
                    sonos_player.soco.add_to_queue(plst)
                    sonos_player.soco.play()
                    self.component.value = 'playing'
                    self.component.save()
                except:
                    print(traceback.format_exc(), file=sys.stderr)
                return



from django import forms
from simo.core.forms import BaseComponentForm
from .models import SonosPlayer, SonosPlaylist


class SONOSPlayerConfigForm(BaseComponentForm):
    sonos_device = forms.ModelChoiceField(
        queryset=SonosPlayer.objects.all()
    )



from django.contrib import admin
from .models import SonosPlayer, SonosPlaylist


class SonosPlaylistInline(admin.TabularInline):
    extra = 0
    model = SonosPlaylist
    fields = 'id', 'title', 'item_id'
    readonly_fields = fields

    def has_add_permission(self, request, obj=None):
        return False




@admin.register(SonosPlayer)
class SonosPlayerAdmin(admin.ModelAdmin):
    list_display = '__str__', 'ip', 'is_alive', 'last_seen', 'is_master'
    search_fields = 'name', 'ip'
    readonly_fields = 'name', 'uid', 'is_master', 'slave_of', 'is_alive', 'last_seen'
    fields = readonly_fields
    inlines = SonosPlaylistInline,

    def has_add_permission(self, request):
        return False

    def is_master(self, obj):
        return not obj.slave_of
    is_master.boolean = True





# Generated by Django 3.2.9 on 2023-12-18 12:18

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('simo_sonos', '0002_rename_master_sonosplayer_slave_of'),
    ]

    operations = [
        migrations.AlterField(
            model_name='sonosplaylist',
            name='item_id',
            field=models.CharField(max_length=200),
        ),
    ]



import time
import threading
from datetime import timedelta
from django.utils import timezone
from soco import discover, SoCo
from soco.snapshot import Snapshot
from simo.core.gateways import BaseObjectCommandsGatewayHandler
from simo.core.forms import BaseGatewayForm
from simo.core.models import Component
from simo.core.utils.helpers import get_self_ip
from simo.multimedia.models import Sound
from .models import SonosPlayer, SonosPlaylist
from .models import SonosPlayer
from .utils import get_sec


class SONOSGatewayHandler(BaseObjectCommandsGatewayHandler):
    name = "SONOS"
    config_form = BaseGatewayForm

    periodic_tasks = (
        ('periodic_players_discovery', 60 * 10),
        ('watch_players', 1)
    )

    playing_alerts = {}
    watch_second = 0

    def perform_value_send(self, component, value):
        sonos_player = SonosPlayer.objects.get(id=component.config['sonos_device'])
        if value in (
            'play', 'pause', 'stop', 'next', 'previous',
        ):
            getattr(sonos_player.soco, value)()
        elif isinstance(value, dict):
            if 'seek' in value:
                sonos_player.soco.seek(timedelta(seconds=value['seek']))
            elif 'set_volume' in value:
                sonos_player.soco.volume = value['set_volume']
            elif 'shuffle' in value:
                sonos_player.soco.shuffle = value['shuffle']
            elif 'loop' in value:
                sonos_player.soco.repeat = value['loop']
            elif 'play_from_library' in value:
                if value['play_from_library'].get('type') != 'sonos_playlist':
                    return
                playlist = SonosPlaylist.objects.filter(
                    id=value['play_from_library'].get('id', 0)
                ).first()
                if not playlist:
                    return
                sonos_player.play_playlist(playlist)
            elif 'play_uri' in value:
                if value.get('volume') != None:
                    sonos_player.soco.volume = value['volume']
                sonos_player.soco.play_uri(value['play_uri'])
            elif 'alert' in value:
                try:
                    sound_id = int(value['alert'])
                except:
                    uri = value['alert']
                    length = None
                else:
                    sound = Sound.objects.get(pk=sound_id)
                    length = sound.length
                    uri = f"http://{get_self_ip()}{sound.get_absolute_url()}"

                threading.Thread(
                    target=self.play_alert, args=(
                        sonos_player, uri, length, value.get('volume')
                    )
                ).start()

        self.comp_state_update(component)


    def play_alert(self, sonos_player, uri, length, volume):
        if sonos_player.id not in self.playing_alerts:
            self.playing_alerts[sonos_player.id] = {
                'snap': Snapshot(sonos_player.soco),
                'uri': uri, 'timestamp': time.time()
            }
            self.playing_alerts[sonos_player.id]['snap'].snapshot()
        else:
            self.playing_alerts[sonos_player.id].update({
                'uri': uri, 'timestamp': time.time()
            })

        start_timestamp = self.playing_alerts[sonos_player.id]['timestamp']

        if volume != None:
            sonos_player.soco.volume = volume

        print("Play alert from URI: ", uri)
        sonos_player.soco.clear_queue()
        sonos_player.soco.play_uri(uri)
        if length != None:
            if length > 60:
                length = 60
            time.sleep(length)
        else:
            for i in range(60):
                time.sleep(1)
                status = sonos_player.soco.get_current_transport_info()
                if status.get(
                    'current_transport_state', 'PLAYING'
                ) != 'PLAYING':
                    break

        if self.playing_alerts[sonos_player.id]['timestamp'] == start_timestamp \
        and self.playing_alerts[sonos_player.id]['uri'] == uri:

            print("Restore original")
            snap = self.playing_alerts[sonos_player.id]['snap']
            try:
                if snap.is_coordinator:
                    snap._restore_coordinator()
            finally:
                snap.device.mute = snap.mute
                snap.device.bass = snap.bass
                snap.device.treble = snap.treble
                snap.device.loudness = snap.loudness
                snap.device.volume = 0
                snap.device.ramp_to_volume(
                    snap.volume, ramp_type='AUTOPLAY_RAMP_TYPE'
                )

            # Now everything is set, see if we need to be playing, stopped
            # or paused ( only for coordinators)
            if snap.is_coordinator:
                if snap.transport_state == "PLAYING":
                    snap.device.play()
                elif snap.transport_state == "STOPPED":
                    snap.device.stop()

            del self.playing_alerts[sonos_player.id]

    def periodic_players_discovery(self):
        # Perform sonos players discovery and state check
        # of non playing players every 10 minutes
        self.discover_sonos_players()
        for comp in Component.objects.filter(
            gateway=self.gateway_instance, base_type='audio-player',
        ).exclude(value='playing'):
            self.comp_state_update(comp)

    def watch_players(self):
        for comp in Component.objects.filter(
            gateway=self.gateway_instance, base_type='audio-player',
            value='playing'
        ):
            self.comp_state_update(comp)

        # Check other players every 60 seconds, just in case...
        if not self.watch_second % 60:
            self.watch_second = 0
            for comp in Component.objects.filter(
                gateway=self.gateway_instance, base_type='audio-player',
            ).exclude(value='playing'):
                self.comp_state_update(comp)
        else:
            self.watch_second += 1

    def discover_sonos_players(self):
        print("Discover SONOS players.")

        discovered_players = []
        sonos_devices = list(discover(allow_network_scan=True))
        for sonos in sonos_devices:
            if sonos.group.coordinator.uid != sonos.uid:
                # Skip slave speakers save group masters only.
                continue
            player, new = SonosPlayer.objects.update_or_create(
                uid=sonos.uid, defaults={
                    'name': sonos.player_name, 'ip': sonos.ip_address,
                    'last_seen': timezone.now(), 'is_alive': True
                }
            )
            discovered_players.append(player)
            if new:
                print(f"New player - {player} - was found!")
            else:
                print(f"{player} - rediscovered.")

        missing_players = SonosPlayer.objects.exclude(
            id__in=[p.id for p in discovered_players]
        )
        if missing_players:
            print("Let's manually check the missing ones!")
            for missing_player in missing_players:
                try:
                    sonos = SoCo(missing_player.ip)
                    sonos.get_speaker_info()
                except:
                    print(
                        f"{missing_player} - still not available at {missing_player.ip}")
                    missing_player.is_alive = False
                    missing_player.save()
                else:
                    sonos_devices.append(sonos)
                    player, new = SonosPlayer.objects.update_or_create(
                        uid=sonos.uid, defaults={
                            'name': sonos.player_name, 'ip': sonos.ip_address,
                            'last_seen': timezone.now(), 'is_alive': True
                        }
                    )
                    if new:
                        print(f"New player - {player} - was found!")
                    else:
                        print(f"{player} - rediscovered.")

        print(
            "Figure out who's da boss "
            "and SONOS playlists that are available."
        )
        for sonos in sonos_devices:
            player = SonosPlayer.objects.get(uid=sonos.uid)
            if sonos.group.coordinator.uid == sonos.uid:
                player.slave_of = None
            else:
                player.slave_of = SonosPlayer.objects.filter(
                    uid=sonos.group.coordinator.uid
                ).first()
            player.save()

            playlists = []

            for pls in sonos.get_sonos_playlists():
                playlist, new = SonosPlaylist.objects.update_or_create(
                    item_id=pls.item_id,
                    player=SonosPlayer.objects.get(uid=sonos.uid),
                    defaults={'title': pls.title}
                )
                playlists.append(playlist)

            if playlists:
                for comp in Component.objects.filter(
                    gateway=self.gateway_instance, base_type='audio-player'
                ):
                    comp.meta['library'] = [
                        {'type': 'sonos_playlist',
                         'id': pls.id, 'title': pls.title}
                        for pls in playlists
                    ]
                    comp.save()

    def comp_state_update(self, sonos_component):
        print(f"Check {sonos_component} state!")
        sonos_player = SonosPlayer.objects.get(
            id=sonos_component.config['sonos_device']
        )
        try:
            status = sonos_player.soco.get_current_transport_info()
        except:
            sonos_component.value = 'stopped'
            sonos_component.alive = False
            sonos_component.save()
            return

        state_map = {
            'PLAYING': 'playing',
            'PAUSED_PLAYBACK': 'paused',
            'STOPPED': 'stopped'
        }
        sonos_component.value = state_map.get(
            status.get('current_transport_state', 'STOPPED'),
            'STOPPED'
        )

        try:
            info = sonos_player.get_current_track_info()
        except:
            sonos_component.alive = False
            sonos_component.save()
            return

        sonos_component.meta.update({
            'title': info['title'],
            'duration': get_sec(info['duration']),
            'position': get_sec(info['position']),
            'volume': sonos_player.soco.volume,
            'shuffle': sonos_player.soco.shuffle,
            'loop': sonos_player.soco.repeat,
        })

        sonos_component.save()



from django.db import models
from django.db.models.signals import post_save
from django.dispatch import receiver
from dirtyfields import DirtyFieldsMixin
from soco import SoCo


class SonosPlayer(DirtyFieldsMixin, models.Model):
    slave_of = models.ForeignKey(
        'SonosPlayer', null=True, blank=True, on_delete=models.SET_NULL,
        related_name='slaves'
    )
    uid = models.CharField(max_length=200, db_index=True, unique=True)
    name = models.CharField(max_length=200, db_index=True)
    ip = models.GenericIPAddressField()
    last_seen = models.DateTimeField(auto_now_add=True)
    is_alive = models.BooleanField(default=True)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.soco = SoCo(self.ip)

    def __str__(self):
        return f'{self.name} ({self.uid}) '


@receiver(post_save, sender=SonosPlayer)
def update_related_components(sender, instance, created, **kwargs):
    if created:
        return
    if 'is_alive' not in instance.get_dirty_fields():
        return
    from simo_sonos.gateways import SONOSGatewayHandler
    from simo.core.models import Component
    for component in Component.objects.filter(
        gateway__type=SONOSGatewayHandler.uid, base_type='audio-player',
    ):
        if component.config.get('sonos_device') != instance.id:
            continue
        component.alive = instance.is_alive
        component.save()


class SonosPlaylist(models.Model):
    player = models.ForeignKey(SonosPlayer, on_delete=models.CASCADE)
    title = models.CharField(max_length=200, db_index=True)
    item_id = models.CharField(max_length=200)

    class Meta:
        unique_together = 'player', 'item_id'

    def __str__(self):
        return f"{self.title} on {self.player}"
